#ifndef	TRUE
#define	TRUE	1
#endif
#ifndef	FALSE
#define	FALSE	0
#endif

#define	CONFIG_FILE	"/usr/local/etc/nrbroadcast"

#define	NODES_FILE	"/proc/net/nr_nodes"
#define	NEIGH_FILE	"/proc/net/nr_neigh"

#define	NETROM_PID	0xCF

struct port_struct {
	char *device;
	int  port;
	int  paclen;
	int  minimum_obs;
	int  default_qual;
	int  worst_qual;
	int  verbose;
};

extern struct port_struct port_list[];

extern int port_count;
extern int debug;

extern ax25_address my_call;
extern ax25_address node_call;

/* In netromd.c */
extern int  ax25cmp(ax25_address *, ax25_address *);
extern char *ax2asc(ax25_address *);

/* In netromt.c */
extern void transmit_nodes(int);

/* In netromr.c */
extern void receive_nodes(unsigned char *, int, ax25_address *, int);
