/* In convert.c */
extern int  convert_call(char *, struct full_sockaddr_ax25 *);
extern int  convert_call_entry(char *, char *);
extern int  convert_call_arglist(char ** ,struct full_sockaddr_ax25 *);

/* In config.c */
extern int  ax25_config_load_ports(void);
extern int  ax25_config_num_ports(void);
extern char *ax25_config_get_addr(int);
extern char *ax25_config_get_dev(int);
extern int  ax25_config_get_window(int);
extern int  ax25_config_get_paclen(int);
extern int  ax25_config_get_baud(int);
extern char *ax25_config_get_desc(int);

extern int  nr_config_load_ports(void);
extern int  nr_config_num_ports(void);
extern char *nr_config_get_addr(int);
extern char *nr_config_get_dev(int);
extern char *nr_config_get_alias(int);
extern int  nr_config_get_paclen(int);
extern char *nr_config_get_desc(int);
